package com.qmth.distributed.print.business.service.impl;

import cn.hutool.core.date.DateUtil;
import com.alibaba.fastjson.JSONObject;
import com.baomidou.mybatisplus.core.conditions.query.QueryWrapper;
import com.baomidou.mybatisplus.core.conditions.update.UpdateWrapper;
import com.qmth.boot.api.exception.ApiException;
import com.qmth.distributed.print.business.bean.dto.ExamPaperObjectiveStructureDto;
import com.qmth.distributed.print.business.bean.dto.ExamPaperSubjectiveStructureDto;
import com.qmth.distributed.print.business.bean.dto.SyncExamCardDto;
import com.qmth.distributed.print.business.bean.dto.SyncExamStudentDto;
import com.qmth.distributed.print.business.bean.params.SyncDataParam;
import com.qmth.distributed.print.business.entity.*;
import com.qmth.distributed.print.business.enums.ExamPaperStructureStatusEnum;
import com.qmth.distributed.print.business.enums.PrintPlanStatusEnum;
import com.qmth.distributed.print.business.service.*;
import com.qmth.teachcloud.common.bean.dto.SyncStructureData;
import com.qmth.teachcloud.common.contant.SystemConstant;
import com.qmth.teachcloud.common.entity.BasicAttachment;
import com.qmth.teachcloud.common.entity.SysConfig;
import com.qmth.teachcloud.common.entity.SysUser;
import com.qmth.teachcloud.common.enums.ExceptionResultEnum;
import com.qmth.teachcloud.common.enums.SyncFileTypeEnum;
import com.qmth.teachcloud.common.enums.TaskResultEnum;
import com.qmth.teachcloud.common.enums.TaskStatusEnum;
import com.qmth.teachcloud.common.enums.userPush.SpecialPrivilegeEnum;
import com.qmth.teachcloud.common.service.BasicAttachmentService;
import com.qmth.teachcloud.common.service.SysConfigService;
import com.qmth.teachcloud.common.service.SysUserService;
import com.qmth.teachcloud.common.service.TeachcloudCommonService;
import com.qmth.teachcloud.common.sync.StmmsUtils;
import com.qmth.teachcloud.common.util.ServletUtil;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.util.CollectionUtils;

import java.io.*;
import java.util.*;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.stream.Collectors;

/**
 * @Date: 2021/5/20.
 */
@Service
public class DataSyncServiceImpl implements DataSyncService {

    private final static Logger log = LoggerFactory.getLogger(DataSyncServiceImpl.class);

    @Autowired
    private ExamPrintPlanService examPrintPlanService;

    @Autowired
    private TBSyncTaskService tbSyncTaskService;

    @Autowired
    private ExamStudentService examStudentService;

    @Autowired
    private ExamDetailService examDetailService;

    @Autowired
    private ExamCardService examCardService;

    @Autowired
    private ExamPaperStructureService examPaperStructureService;

    @Autowired
    private SysConfigService sysConfigService;

    @Autowired
    private TeachcloudCommonService teachcloudCommonService;

    @Autowired
    private BasicAttachmentService basicAttachmentService;

    @Autowired
    private SysUserService sysUserService;

    @Autowired
    StmmsUtils stmmsUtils;

    private ExecutorService executors = Executors.newFixedThreadPool(5);

    private static final String DATE_FORMAT = "yyyy-MM-dd HH:mm:ss";

    /**
     * 定时任务批量同步
     */
    @Override
    public void syncToCloudReview() {
        // 查询可同步计划（同步状态为空：未同步，false：同步失败）
        QueryWrapper<ExamPrintPlan> queryWrapper = new QueryWrapper<>();
        queryWrapper.lambda().eq(ExamPrintPlan::getStatus, PrintPlanStatusEnum.END)
                .and(q -> q.isNull(ExamPrintPlan::getSyncStatus).or().eq(ExamPrintPlan::getSyncStatus, false));
        List<ExamPrintPlan> examPrintPlans = examPrintPlanService.list(queryWrapper);
        if (!CollectionUtils.isEmpty(examPrintPlans)) {
            for (ExamPrintPlan examPrintPlan : examPrintPlans) {
                executors.execute(syncData(examPrintPlan, null, examPrintPlan.getName()));
            }
        }
    }

    /**
     * 单个同步
     *
     * @param printPlanId
     * @param thirdRelateId
     */
    @Override
    public void syncDataCloud(Long printPlanId, Long thirdRelateId) {
        ExamPrintPlan examPrintPlan = examPrintPlanService.getById(printPlanId);
        if (examPrintPlan == null) {
            throw ExceptionResultEnum.ERROR.exception("印刷计划数据异常");
        }
        if (!PrintPlanStatusEnum.END.equals(examPrintPlan.getStatus()) && !PrintPlanStatusEnum.PRINT_FINISH.equals(examPrintPlan.getStatus())) {
            throw ExceptionResultEnum.ERROR.exception("印刷计划未打印完成或者未结束，不能同步数据");
        }
        Long relateId = Objects.isNull(thirdRelateId) ? examPrintPlan.getThirdRelateId() : thirdRelateId;
        String thirdRelateName = Objects.isNull(examPrintPlan.getThirdRelateId()) ? examPrintPlan.getName() : examPrintPlan.getThirdRelateName();
        executors.execute(syncData(examPrintPlan, relateId, thirdRelateName));

    }

    @Override
    public void syncDataMerge(SyncDataParam syncDataParam) {
        List<Long> printPlans = syncDataParam.getList();
        if (CollectionUtils.isEmpty(printPlans)) {
            throw ExceptionResultEnum.ERROR.exception("请选择需要合并的印刷任务");
        }

        if (Objects.isNull(syncDataParam.getThirdRelateId()) && Objects.isNull(syncDataParam.getThirdRelateName())) {
            throw ExceptionResultEnum.ERROR.exception("考试ID、考试名称至少填一个");
        }

        // todo 校验印刷计划是否完成
        if (syncDataParam.getThirdRelateId() == null) {
            String time = String.valueOf(System.currentTimeMillis());
            QueryWrapper<ExamPrintPlan> printPlanQueryWrapper = new QueryWrapper<>();
            printPlanQueryWrapper.lambda().in(ExamPrintPlan::getId, syncDataParam.getList());
            List<ExamPrintPlan> examPrintPlans = examPrintPlanService.list(printPlanQueryWrapper);
            ExamPrintPlan examPrintPlan = examPrintPlans.stream().max(Comparator.comparingLong(ExamPrintPlan::getExamEndTime)).get();
            String examTime = DateUtil.format(new Date(examPrintPlan.getExamEndTime()), DATE_FORMAT);
            syncDataParam.setThirdRelateId(stmmsUtils.syncExam(time, syncDataParam.getThirdRelateName(), examTime));
        }

        for (Long printPlanId : printPlans) {
            ExamPrintPlan printPlan = examPrintPlanService.getById(printPlanId);
            executors.execute(syncData(printPlan, syncDataParam.getThirdRelateId(), syncDataParam.getThirdRelateName()));
        }
    }

    @Override
    public void uploadStructure(ExamPaperStructure examPaperStructure) {
        // 开始同步
        examPaperStructureService.updateStatusById(examPaperStructure.getId(), ExamPaperStructureStatusEnum.START_SYNC);
        ExamPaperStructureStatusEnum status = ExamPaperStructureStatusEnum.UPLOAD_FINISH;
        try {
            String paperAnswer = examPaperStructure.getPaperAnswer();
            if (StringUtils.isBlank(paperAnswer)) {
                throw ExceptionResultEnum.ERROR.exception("没有找到标答文件");
            }

            JSONObject paperAnswerJson = JSONObject.parseObject(paperAnswer);
            if (!paperAnswerJson.containsKey("paperType") || !paperAnswerJson.containsKey("paper") || !paperAnswerJson.containsKey("answer")) {
                throw ExceptionResultEnum.ERROR.exception("没有找到标答文件");
            }
            String paperType = paperAnswerJson.getString("paperType");
            String paper = paperAnswerJson.getString("paper");
            String answer = paperAnswerJson.getString("answer");
            if (StringUtils.isAnyBlank(paperType, paper, answer)) {
                throw ExceptionResultEnum.ERROR.exception("没有找到标答文件");
            }

            // 同步试卷文件
            File paperFile = downFileFromFss(paper, paperType, SyncFileTypeEnum.PAPER);
            boolean syncPaperFile = stmmsUtils.syncFile(String.valueOf(examPaperStructure.getThirdRelateId()), examPaperStructure.getPaperNumber(), SyncFileTypeEnum.PAPER, paperFile);
            if (syncPaperFile) {
                status = ExamPaperStructureStatusEnum.PAPER_FINISH;
            }

            // 同步标答文件
            File answerFile = downFileFromFss(paper, paperType, SyncFileTypeEnum.PAPER);
            boolean syncAnswerFile = stmmsUtils.syncFile(String.valueOf(examPaperStructure.getThirdRelateId()), examPaperStructure.getPaperNumber(), SyncFileTypeEnum.ANSWER, answerFile);
            if (syncAnswerFile) {
                status = ExamPaperStructureStatusEnum.ANSWER_FINISH;
            }
            // 同步客观题
            String objectiveStructure = examPaperStructure.getObjectiveStructure();
            if (StringUtils.isBlank(objectiveStructure)) {
                throw ExceptionResultEnum.ERROR.exception("没有找到客观题数据");
            }
            JSONObject objectiveJson = JSONObject.parseObject(objectiveStructure);
            if (!objectiveJson.containsKey("paperType") || !objectiveJson.containsKey("content")) {
                throw ExceptionResultEnum.ERROR.exception("客观题文件未上传");
            }
            String objectivePaperType = objectiveJson.getString("paperType");
            String objectiveContent = objectiveJson.getString("content");
            if (StringUtils.isAnyBlank(objectivePaperType, objectiveContent)) {
                throw ExceptionResultEnum.ERROR.exception("没有找到客观题数据");
            }
            List<ExamPaperObjectiveStructureDto> objectiveStructureDtos = JSONObject.parseArray(objectiveContent, ExamPaperObjectiveStructureDto.class);
            List<SyncStructureData> syncObjectiveStructureDatas = objectiveStructureDtos.stream().map(m -> {
                SyncStructureData syncStructureData = new SyncStructureData();
                syncStructureData.setMainNumber(Integer.valueOf(m.getMainNumber()));
                syncStructureData.setSubNumber(m.getSubNumber());
                syncStructureData.setMainTitle(m.getMainName());
                syncStructureData.setTotalScore(Double.valueOf(m.getScore()));
                syncStructureData.setAnswer(m.getAnswer());
                return syncStructureData;
            }).collect(Collectors.toList());
            boolean syncObjectiveStructure = stmmsUtils.syncPaperStructure(String.valueOf(examPaperStructure.getThirdRelateId()), examPaperStructure.getPaperNumber(), true, objectivePaperType, syncObjectiveStructureDatas);
            if (syncObjectiveStructure) {
                status = ExamPaperStructureStatusEnum.OBJECTIVE_FINISH;
            }
            // 同步主观题
            String subjectiveStructure = examPaperStructure.getSubjectiveStructure();
            if (StringUtils.isBlank(subjectiveStructure)) {
                throw ExceptionResultEnum.ERROR.exception("没有找到主观题数据");
            }
            JSONObject subjectiveJson = JSONObject.parseObject(subjectiveStructure);
            if (!subjectiveJson.containsKey("paperType") || !subjectiveJson.containsKey("content")) {
                throw ExceptionResultEnum.ERROR.exception("主观题文件未上传");
            }
            String subjectivePaperType = subjectiveJson.getString("paperType");
            String subjectieContent = subjectiveJson.getString("content");
            if (StringUtils.isAnyBlank(subjectivePaperType, subjectieContent)) {
                throw ExceptionResultEnum.ERROR.exception("没有找到主观题数据");
            }
            List<ExamPaperSubjectiveStructureDto> subjectiveStructureDtos = JSONObject.parseArray(subjectieContent, ExamPaperSubjectiveStructureDto.class);
            List<SyncStructureData> syncSubjectiveStructureDatas = subjectiveStructureDtos.stream().map(m -> {
                SyncStructureData syncStructureData = new SyncStructureData();
                syncStructureData.setMainNumber(Integer.valueOf(m.getMainNumber()));
                syncStructureData.setSubNumber(m.getSubNumber());
                syncStructureData.setMainTitle(m.getMainName());
                syncStructureData.setTotalScore(Double.valueOf(m.getScore()));
                return syncStructureData;
            }).collect(Collectors.toList());
            boolean syncSubjectiveStructure = stmmsUtils.syncPaperStructure(String.valueOf(examPaperStructure.getThirdRelateId()), examPaperStructure.getPaperNumber(), true, subjectivePaperType, syncSubjectiveStructureDatas);
            if (syncSubjectiveStructure) {
                status = ExamPaperStructureStatusEnum.SUBJECTIVE_FINISH;
            }
        } catch (ApiException e) {
            throw ExceptionResultEnum.ERROR.exception("试卷结构同步失败：" + e.getMessage());
        } finally {
            examPaperStructureService.updateStatusById(examPaperStructure.getId(), status);
        }
    }

    private TimerTask syncData(ExamPrintPlan examPrintPlan, Long thirdRelateId, String thirdRelateName) {
        return new TimerTask() {
            @Override
            public void run() {
                SysConfig sysConfig = sysConfigService.getByKey("sys.sync.enable");
                if (sysConfig != null && "true".equals(sysConfig.getConfigValue())) {
                    TBSyncTask syncTask = tbSyncTaskService.saveTask(examPrintPlan);
                    doSyncCore(examPrintPlan, thirdRelateId, thirdRelateName, syncTask);
                }
            }
        };
    }

    /**
     * 同步核心方法
     *
     * @param thirdRelateName
     * @param syncTask
     */
    public void doSyncCore(ExamPrintPlan examPrintPlan, Long thirdRelateId, String thirdRelateName, TBSyncTask syncTask) {
        UpdateWrapper<ExamPrintPlan> updateWrapper = new UpdateWrapper<>();
        try {
            if (!PrintPlanStatusEnum.END.equals(examPrintPlan.getStatus()) && !PrintPlanStatusEnum.PRINT_FINISH.equals(examPrintPlan.getStatus())) {
                throw ExceptionResultEnum.ERROR.exception("印刷计划未打印完成或者未结束，不能同步数据");
            }
            UpdateWrapper<TBSyncTask> tbSyncTaskUpdateWrapper = new UpdateWrapper<>();
            tbSyncTaskUpdateWrapper.lambda().set(TBSyncTask::getStatus, TaskStatusEnum.RUNNING).eq(TBSyncTask::getId, syncTask.getId());
            tbSyncTaskService.update(tbSyncTaskUpdateWrapper);

            // 同步计划 -> 对应云阅卷考试
            thirdRelateId = saveExam(examPrintPlan, thirdRelateId);
            syncTask.setThirdRelateId(thirdRelateId);
            // 考试同步成功，才能同步考生和题卡
            if (Objects.nonNull(thirdRelateId)) {
                updateWrapper.lambda().set(ExamPrintPlan::getThirdRelateId, thirdRelateId).set(ExamPrintPlan::getThirdRelateName, thirdRelateName);
                saveStudent(examPrintPlan.getId(), thirdRelateId);
                cardUpload(examPrintPlan.getId(), thirdRelateId);
            } else {
                throw ExceptionResultEnum.ERROR.exception("无法获取云阅卷考试，同步数据失败");
            }

            // 更新计划状态
            updateWrapper.lambda().set(ExamPrintPlan::getSyncStatus, true);

            // 更新日志表
            syncTask.setResult(TaskResultEnum.SUCCESS);
            syncTask.setErrorMessage(null);
        } catch (Exception e) {
            syncTask.setResult(TaskResultEnum.ERROR);
            syncTask.setErrorMessage(e.getMessage());
            updateWrapper.lambda().set(ExamPrintPlan::getSyncStatus, false);
        } finally {
            syncTask.setStatus(TaskStatusEnum.FINISH);
            tbSyncTaskService.saveOrUpdate(syncTask);
            updateWrapper.lambda().eq(ExamPrintPlan::getId, examPrintPlan.getId());
            examPrintPlanService.update(updateWrapper);
        }

    }

    /**
     * 创建考试
     *
     * @param examPrintPlan
     * @return
     */
    public Long saveExam(ExamPrintPlan examPrintPlan, Long thirdRelateId) {
        try {
            if (Objects.isNull(thirdRelateId)) {
                String code = String.valueOf(examPrintPlan.getId());
                String name = examPrintPlan.getName();
                String examTime = DateUtil.format(new Date(examPrintPlan.getExamEndTime()), DATE_FORMAT);
                thirdRelateId = stmmsUtils.syncExam(code, name, examTime);
            }
            return thirdRelateId;
        } catch (Exception e) {
            throw ExceptionResultEnum.ERROR.exception(e.getMessage());
        }
    }

    /**
     * 新增考生
     *
     * @return
     */
    public void saveStudent(Long examPrintPlanId, Long thirdRelateId) {
        List<SyncExamStudentDto> examStudents = examStudentService.listStudentByPrintPlanIdAndSyncStatus(examPrintPlanId);
        for (SyncExamStudentDto examStudent : examStudents) {
            try {
                //参数
                String examId = String.valueOf(thirdRelateId);
                String examNumber = examStudent.getTicketNumber();
                String studentCode = examStudent.getStudentCode();
                String name = examStudent.getStudentName();
                String college = "无";
                String className = getClassName(examStudent.getExtendFields());
                String teacher = "无";
                String subjectCode = examStudent.getPaperNumber(); // 取试卷编号
                String subjectName = examStudent.getCourseName();
                boolean syncStudent = stmmsUtils.syncStudent(examId, examNumber, studentCode, name, college, className, teacher, subjectCode, subjectName, null, null, null, null);

                if (syncStudent) {
                    UpdateWrapper<ExamStudent> updateWrapper = new UpdateWrapper<>();
                    updateWrapper.lambda().set(ExamStudent::getSyncStatus, true).eq(ExamStudent::getId, examStudent.getId());
                    examStudentService.update(updateWrapper);
                }
            } catch (Exception e) {
                throw ExceptionResultEnum.ERROR.exception(e.getMessage());
            }
        }
    }

    /**
     * 同步题卡
     *
     * @return
     */
    public void cardUpload(Long examPrintPlanId, Long thirdRelateId) {
        List<ExamDetailCourse> examDetailCourses = examDetailService.listSyncPaperNumberByPrintPlanId(examPrintPlanId);
        if (CollectionUtils.isEmpty(examDetailCourses)) {
            return;
        }
        for (ExamDetailCourse examDetailCours : examDetailCourses) {
            List<SyncExamCardDto> syncExamCardDtos = examCardService.listSyncCardByCourseCodeAndPaperNumber(examDetailCours.getSchoolId(), examDetailCours.getCourseCode(), examDetailCours.getPaperNumber());
            if (CollectionUtils.isEmpty(syncExamCardDtos)) {
                return;
            }

            if (syncExamCardDtos.size() > 1) {
                log.info("数据异常，通过学校：{},课程代码：{},试卷编号：{}查出{}个题卡数据", examDetailCours.getSchoolId(), examDetailCours.getCourseCode(), examDetailCours.getPaperNumber(), syncExamCardDtos.size());
                return;
            }

            SyncExamCardDto syncExamCardDto = syncExamCardDtos.get(0);
            if (StringUtils.isNotEmpty(syncExamCardDto.getContent())) {
                //生成json文件
                File file = null;
                try {
                    // 文件临时目录
                    String filePath = getTempDir(SyncFileTypeEnum.CARD);
                    file = createJsonFile(filePath, syncExamCardDto.getContent());
                    if (file.exists()) {
                        boolean uploadCard = stmmsUtils.syncFile(String.valueOf(thirdRelateId), syncExamCardDto.getPaperNumber(), SyncFileTypeEnum.CARD, file);
                        if (uploadCard) {
                            UpdateWrapper<ExamCard> updateWrapper = new UpdateWrapper<>();
                            updateWrapper.lambda().set(ExamCard::getSyncStatus, true).eq(ExamCard::getId, syncExamCardDto.getId());
                            examCardService.update(updateWrapper);
                        }
                    }
                } catch (Exception e) {
                    throw ExceptionResultEnum.ERROR.exception(e.getMessage());
                } finally {
                    if (file != null && file.exists()) {
                        file.delete();
                    }
                }
            }

        }
    }

    @Override
    public Map<String, Object> markerLoginInfo() {
        try {
            SysUser sysUser = (SysUser) ServletUtil.getRequestUser();
            SpecialPrivilegeEnum userSpecialPrivilege = sysUserService.findUserSpecialPrivilegeByUserId(sysUser.getId());
            if (SpecialPrivilegeEnum.UNIDENTIFIED.equals(userSpecialPrivilege)
                    || SpecialPrivilegeEnum.SUBJECT_HEADER.equals(userSpecialPrivilege)) {
                throw ExceptionResultEnum.ERROR.exception("该用户没有评卷员角色，无法登录");
            }
            return stmmsUtils.markLogin(sysUser);
        } catch (Exception e) {
            throw ExceptionResultEnum.ERROR.exception(e.getMessage());
        }
    }

    @Override
    public Map<String, Object> markerLeaderLoginInfo() {
        try {
            SysUser sysUser = (SysUser) ServletUtil.getRequestUser();
            SpecialPrivilegeEnum userSpecialPrivilege = sysUserService.findUserSpecialPrivilegeByUserId(sysUser.getId());
            if (SpecialPrivilegeEnum.UNIDENTIFIED.equals(userSpecialPrivilege)
                    || SpecialPrivilegeEnum.MARKER.equals(userSpecialPrivilege)) {
                throw ExceptionResultEnum.ERROR.exception("该用户没有科组长角色，无法登录");
            }
            return stmmsUtils.markLeaderLogin(sysUser);
        } catch (Exception e) {
            throw ExceptionResultEnum.ERROR.exception(e.getMessage());
        }
    }

    /**
     * 上传文件临时目录
     * @param type
     * @return
     */
    private String getTempDir(SyncFileTypeEnum type) {
        return SystemConstant.TEMP_FILES_DIR + File.separator + "upload-temp" + File.separator + type.name().toLowerCase() + File.separator + System.currentTimeMillis();
    }

    private String getClassName(String extendCloumn) {
        List<Map> mapList = JSONObject.parseArray(extendCloumn, Map.class);
        for (Map map : mapList) {
            if (Objects.equals("className", map.get("code").toString())) {
                return map.get("value").toString();
            }
        }
        return "无";
    }

    /*
     * 生成文件
     * @param file 文件路径+文件名称
     * @param conent 要生成的文件内容
     */
    public static File createJsonFile(String url, String conent) {

        File file = new File(url);
        if (!file.exists()) {
            file.mkdirs();
        }
        BufferedWriter out = null;
        file = new File(file, UUID.randomUUID().toString() + ".json");
        try {
            file.createNewFile();
            out = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(file, true)));
            out.write(conent);
        } catch (Exception e) {
        } finally {
            try {
                out.close();
            } catch (IOException e) {
            }
        }
        return file;
    }

    /**
     * 下载附件临时文件
     *
     * @param attachmentId
     * @param name
     * @param type
     * @return
     */
    private File downFileFromFss(String attachmentId, String name, SyncFileTypeEnum type) {
        try {
            if (Objects.nonNull(attachmentId)) {
                BasicAttachment attachment = basicAttachmentService.getById(attachmentId);
                if (attachment == null) {
                    throw ExceptionResultEnum.ERROR.exception("附件数据异常");
                }
                String fileName = attachment.getName() + "-" + name + attachment.getType();
                File file = teachcloudCommonService.copyFile(getTempDir(type), fileName, attachment);
                return file;
            }
        } catch (Exception e) {
            throw ExceptionResultEnum.ERROR.exception(e.getMessage());
        }
        return null;
    }
}
